@@ -37,6 +37,11 @@ func (lex *lexer) Init(s string) {
 	lex.sTail = s
 }
 
+func (lex *lexer) PushBack(currToken, sHead string) {
+	lex.Token = currToken
+	lex.sTail = sHead + lex.sTail
+}
+
 func (lex *lexer) Next() error {
 	if lex.err != nil {
 		return lex.err
@@ -117,6 +122,14 @@ again:
 		}
 		goto tokenFoundLabel
 	}
+	if strings.HasPrefix(s, "$__interval") {
+		lex.sTail = s[len("$__interval"):]
+		return "$__interval", nil
+	}
+	if strings.HasPrefix(s, "$__rate_interval") {
+		lex.sTail = s[len("$__rate_interval"):]
+		return "$__interval", nil
+	}
 	return "", fmt.Errorf("cannot recognize %q", s)
 
 tokenFoundLabel:
@@ -228,7 +241,7 @@ func scanPositiveNumber(s string) (string, error) {
 		}
 		return s[:i], nil
 	}
-	for i < len(s) && isDecimalChar(s[i]) {
+	for i < len(s) && isDecimalCharOrUnderscore(s[i]) {
 		i++
 	}
 
@@ -253,7 +266,7 @@ func scanPositiveNumber(s string) (string, error) {
 		// Scan fractional part. It cannot be empty.
 		i++
 		j := i
-		for j < len(s) && isDecimalChar(s[j]) {
+		for j < len(s) && isDecimalCharOrUnderscore(s[j]) {
 			j++
 		}
 		i = j
@@ -382,6 +395,29 @@ func unescapeIdent(s string) string {
 	}
 }
 
+func hasEscapedChars(s string) bool {
+	i := 0
+	for i < len(s) {
+		r, size := utf8.DecodeRuneInString(s[i:])
+		if i == 0 && !isFirstIdentChar(r) || i > 0 && !isIdentChar(r) {
+			return true
+		}
+		i += size
+	}
+	return false
+}
+
+func appendQuotedIdent(dst []byte, s string) []byte {
+	dst = utf8.AppendRune(dst, '"')
+	for i := 0; i < len(s); {
+		r, size := utf8.DecodeRuneInString(s[i:])
+		dst = utf8.AppendRune(dst, r)
+		i += size
+	}
+	dst = utf8.AppendRune(dst, '"')
+	return dst
+}
+
 func appendEscapedIdent(dst []byte, s string) []byte {
 	i := 0
 	for i < len(s) {
@@ -396,6 +432,13 @@ func appendEscapedIdent(dst []byte, s string) []byte {
 	return dst
 }
 
+func ifEscapedCharsAppendQuotedIdent(dst []byte, s string) []byte {
+	if hasEscapedChars(s) {
+		return appendQuotedIdent(dst, s)
+	}
+	return appendEscapedIdent(dst, s)
+}
+
 func (lex *lexer) Prev() {
 	lex.nextTokens = append(lex.nextTokens, lex.Token)
 	lex.Token = lex.prevTokens[len(lex.prevTokens)-1]
@@ -491,6 +534,9 @@ func scanSpecialIntegerPrefix(s string) (skipChars int, isHex bool) {
 }
 
 func isPositiveDuration(s string) bool {
+	if s == "$__interval" {
+		return true
+	}
 	n := scanDuration(s)
 	return n == len(s)
 }
@@ -559,12 +605,14 @@ func DurationValue(s string, step int64) (int64, error) {
 }
 
 func parseSingleDuration(s string, step int64) (float64, error) {
+	if s == "$__interval" {
+		return float64(step), nil
+	}
+
 	s = strings.ToLower(s)
 	numPart := s[:len(s)-1]
-	if strings.HasSuffix(numPart, "m") {
-		// Duration in ms
-		numPart = numPart[:len(numPart)-1]
-	}
+	// Strip trailing m if the duration is in ms
+	numPart = strings.TrimSuffix(numPart, "m")
 	f, err := strconv.ParseFloat(numPart, 64)
 	if err != nil {
 		return 0, fmt.Errorf("cannot parse duration %q: %s", s, err)
@@ -572,25 +620,25 @@ func parseSingleDuration(s string, step int64) (float64, error) {
 	var mp float64
 	switch s[len(numPart):] {
 	case "ms":
-		mp = 1e-3
-	case "s":
 		mp = 1
+	case "s":
+		mp = 1000
 	case "m":
-		mp = 60
+		mp = 60 * 1000
 	case "h":
-		mp = 60 * 60
+		mp = 60 * 60 * 1000
 	case "d":
-		mp = 24 * 60 * 60
+		mp = 24 * 60 * 60 * 1000
 	case "w":
-		mp = 7 * 24 * 60 * 60
+		mp = 7 * 24 * 60 * 60 * 1000
 	case "y":
-		mp = 365 * 24 * 60 * 60
+		mp = 365 * 24 * 60 * 60 * 1000
 	case "i":
-		mp = float64(step) / 1e3
+		mp = float64(step)
 	default:
 		return 0, fmt.Errorf("invalid duration suffix in %q", s)
 	}
-	return mp * f * 1e3, nil
+	return mp * f, nil
 }
 
 // scanDuration scans duration, which must start with positive num.
@@ -623,6 +671,9 @@ func scanSingleDuration(s string, canBeNegative bool) int {
 	if s[0] == '-' && canBeNegative {
 		i++
 	}
+	if s[i:] == "$__interval" {
+		return i + len("$__interval")
+	}
 	for i < len(s) && isDecimalChar(s[i]) {
 		i++
 	}
@@ -670,6 +721,10 @@ func isDecimalChar(ch byte) bool {
 	return ch >= '0' && ch <= '9'
 }
 
+func isDecimalCharOrUnderscore(ch byte) bool {
+	return isDecimalChar(ch) || ch == '_'
+}
+
 func isHexChar(ch byte) bool {
 	return isDecimalChar(ch) || ch >= 'a' && ch <= 'f' || ch >= 'A' && ch <= 'F'
 }
